//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "ffl.h"
#include "Spectrum.h"

using namespace std;

int main (int argc, char* argv[]){

  // check the argument
  if(argc!=4){
    cerr<<""<<endl;
    cerr<<"usage:"<<endl;
    cerr<<argv[0]<<" [GPS time] [ffl file] [nsigma]"<<endl; 
    cerr<<""<<endl;
    return 1;
  }

  // hard-coded parameters
  double df = 0.5; // PSD resolution (Hz)
  int sampling_slow = 50; // sampling rate to define a "slow" channel [Hz]
  double dt_frac = 0.4; // time fraction used to compare before/after for slow channels
  double gps_T = 500; // time window [s]

  // timing parameters
  double gps = atof(argv[1]);

  // nsigma
  double nsigma = atof(argv[3]);

  // load ffl file
  string fflfile = (string)argv[2];
  ffl *FFL = new ffl(fflfile);
  if(!FFL->LoadFrameFile(gps)) return 2;

  // get channel list
  vector <string> channels = FFL->GetChannelList();
  int sampling;

  // locals
  int dummy;
  int Np; // number of points in the PSD
  Spectrum *Spec_before = new Spectrum(4,gps_T,32,0); // spectrum
  Spectrum *Spec_after = new Spectrum(4,gps_T,32,0); // spectrum                                                                                                                           
  double *data; // data container
  int vsize; // data size
  TGraph *before;
  TGraph *after;
  double mean_before, mean_after;
  double rms_before, rms_after;
  int n_before, n_after;

  // plots
  GwollumPlot *GP = new GwollumPlot("BA");
  GP->ResizePlot(1800,750);

  // loop over channels
  for(int c=0; c<(int)channels.size(); c++){

    cout<<c<<"/"<<(int)channels.size()<<": "<<channels[c]<<endl;

    // get sampling rate
    sampling=FFL->GetChannelSampling(channels[c],dummy);

    // get data
    data=FFL->GetData(vsize,channels[c],(int)gps-gps_T,(int)gps+gps_T);
    if(vsize!=2*gps_T*sampling) continue;

    // ********************* test for slow channels
    if(sampling<sampling_slow){

      // init
      mean_before = 0.0;
      mean_after = 0.0;
      rms_before = 0.0;
      rms_after = 0.0;
      n_before=0;
      n_after=0;

      // loop over samples
      for(int i=0; i<vsize; i++){
	
	// before
	if(i < (int)(dt_frac*vsize)){
	  mean_before+=data[i];
	  rms_before+=data[i]*data[i];
	  n_before++;
	}

	// after
	else if(i>(int)((1.0-dt_frac)*vsize)){
          mean_after+=data[i];
          rms_after+=data[i]*data[i];
	  n_after++;	
	}

	// skip
	else continue;
      }

      // compute rms
      mean_before/=(double)n_before;
      mean_after/=(double)n_after;
      rms_before = sqrt(rms_before/(double)n_before - mean_before*mean_before);
      rms_after = sqrt(rms_after/(double)n_after - mean_after*mean_after);

      // stability test
      if(mean_after > mean_before + nsigma*rms_before || mean_after < mean_before - nsigma*rms_before)// something changed
        cout<<"----------------------------- Mean value changed: "<<scientific<<mean_before<<" "<<mean_after<<" "<<rms_before<<endl;
      // ADDME: check rms for oscillation

      // finally check the glitch sample
      for(int i=0; i<vsize; i++){
	if(gps>(int)gps-gps_T+(double)i/sampling
	   && gps<(int)gps-gps_T+(double)(i+1)/sampling){// glitch sample
	  if(data[i] > mean_before + nsigma*rms_before || data[i] < mean_before - nsigma*rms_before)
	    cout<<"----------------------------- Glitch value is special"<<endl;
	  break;
	}
      }

    }

    // ********************* test for fast channels
    else{

      // make spectrum
      Np = NextPowerOfTwo((double)sampling/2.0/df);
      if(Np!=Spec_before->GetSpectrumSize()){
	delete Spec_before;
	Spec_before = new Spectrum(Np,gps_T,sampling,0);
        delete Spec_after;
	Spec_after = new Spectrum(Np,gps_T,sampling,0);
      }

      // add data
      if(!Spec_before->AddData(vsize/2,&(data[0])) || !Spec_after->AddData(vsize/2,&(data[vsize/2]))){
	delete data;
	continue;
      }
      delete data;

      // loop over frequencies
      for(int f=0; f<Spec_before->GetSpectrumSize(); f++){
	rms_before = sqrt(Spec_before->GetPowerVariance(f));

	if(Spec_after->GetPower(f) > Spec_before->GetPower(f) + nsigma*rms_before){
	  cout<<"----------------------------- Spectrum changed (+)"<<endl;
	  after = Spec_after->GetASD();
	  after->SetLineColor(2);
	  after->SetLineWidth(2);
	  GP->Draw(after,"AL");
	  before = Spec_before->GetASD();
	  GP->Draw(before,"Lsame");
	  GP->SetLogx(1);
	  GP->SetLogy(1);
	  after->GetHistogram()->SetTitle(channels[c].c_str());
	  after->GetHistogram()->GetXaxis()->SetTitle("Frequency [Hz]");
	  after->GetHistogram()->GetYaxis()->SetTitle("Amplitude spectrum density [Amp/#sqrt{Hz}]");
	  GP->Print("./BA_"+channels[c]+".png");
	  delete before;
	  delete after;
	  break;
	}
	if(Spec_after->GetPower(f) < Spec_before->GetPower(f) - nsigma*rms_before){
	  cout<<"----------------------------- Spectrum changed (-)"<<endl;
	  before = Spec_before->GetASD();
	  before->SetLineWidth(2);
	  GP->Draw(before,"AL");
	  after = Spec_after->GetASD();
          after->SetLineColor(2);
	  GP->Draw(after,"Lsame");
	  GP->SetLogx(1);
	  GP->SetLogy(1);
	  before->GetHistogram()->SetTitle(channels[c].c_str());
	  before->GetHistogram()->GetXaxis()->SetTitle("Frequency [Hz]");
	  before->GetHistogram()->GetYaxis()->SetTitle("Amplitude spectrum density [Amp/#sqrt{Hz}]");
	  GP->Print("./BA_"+channels[c]+".png");
	  delete before;
	  delete after;
	  break;
	}
      }

    }
  }

  delete FFL;
  delete GP;
  return 0;
}

