//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __ffl__
#define __ffl__

#include "ReadAscii.h"
#include "GwollumPlot.h"
#include "Segments.h"
#include "TGraph.h"
#include "FrameL.h"
#include "FrVect.h"

using namespace std;

/**
 * Load frame files and access data in frames.
 * This class was designed to load frame files and provide tools to access the data. The native format for file registration is the FFL format. A FFL file is a text file where each line identifies a frame file. Each line must contained exactly 5 columns:
 * - path to a frame file
 * - starting GPS time of the frame file
 * - duration of the frame file [s]
 * - unused column [0]
 * - unused column [0]
 *
 * The frame files must be time-sorted.
 *
 * The FFL format is the native format of this class. However the lalcache format (LCF) is also supported. In that case the LCF file is first converted to an FFL file before processing.
 *
 * This class is built using the <a href="http://lappweb.in2p3.fr/virgo/FrameL/">Frame Library</a> functions.
 *
 * \author    Florent Robinet
 */
class ffl {

 public:

  /**
   * @name Constructors and destructors
   @{
  */
  /**
   * Constructor of the ffl class.
   * The frame file list should be provided. Two format are supported: the Frame File List (FFL) and the lalcache format (LCF). The FFL format is the native format for this class, if a LCF file is provided, this file is first converted to an FFL.
   * @param aFrameFileList path to frame file list
   * @param aPlotStyle plotting style
   * @param aVerbosity verbosity level
   */
  ffl(const string aFrameFileList, const string aPlotStyle="GWOLLUM", const int aVerbosity=0);

  /**
   * Destructor of the ffl class.
   */
  virtual ~ffl(void);
  /**
     @}
  */

  /**
   * Sets a new name for this.
   * It is possible to assign a specific name to the object. This is useful for unicity (GwollumPlot object).
   * @param aName new object name
   */
  void SetName(const string aName);

  /**
   * Defines a temporary working directory.
   * When working with another file format than the native FFL, a conversion to FFL is required. The converted file is saved in a working directory defined with this function. The directory must exist before calling this function.
   *
   * By default, the current directory is used.
   * @param aTmpDirPath path to the working directory.
   */
  bool DefineTmpDir(const string aTmpDirPath=".");
  
  /**
   * Loads (or re-loads) the frame file list defined in ffl().
   * If the format is different from the native FFL format, the frame file list is first converted into a working FFL file in the tmp directory defined with DefineTmpDir().
   * This function loads the working ffl file. It is possible to configure several trials to load the FFL file.
   * This function can also be used if the frame file list needs to be reloaded (for example if it changed on disk).
   *
   * Note that, by default, the channel information is extracted using the first frame file in the list. To update the channel information, use ExtractChannels().
   * @param[in] gps GPS time where to scan the channel list
   * @param[in] tryN Number of trials to read the FFL file.
   * @param[in] sleepN Number of seconds to wait between retries.
   */
  bool LoadFrameFile(const int gps=0, const int tryN=1, const int sleepN=1);

  /**
   * Returns a data vector.
   * A pointer to the data vector is returned. Data of channel named 'channel_name' is loaded between 2 GPS times. The size of the returned vector is given by 'aSize'. NULL is returned if this function fails and aSize is assigned with a negative value:
   * -  0: the ffl object is corrupted
   * - -1: the requested GPS range does not make sense
   * - -2: the requested GPS range is not compatible with the frame list
   * - -3: the requested channel cannot be loaded.
   * - -4: the data vector cannot be loaded
   * - -5: some samples are corrupted in the data vector
   * - -6: missing frames (see below)
   *
   * Note that data access is retried several times, as configured with LoadFrameFile(), before sending the error code -4.
   *
   * If there are missing frames in the requested time stretch, no data vector is returned and the error code -6 is returned ('val_missing=100'). It is possible to tag the missing samples by a special value given by 'val_missing', which must be different from 100.
   *
   * The user is in charge of deleting the returned vector.
   * @param aSize number of samples in the returned vector
   * @param channel_name channel name
   * @param gps_start starting GPS time
   * @param gps_end ending GPS time
   * @param val_missing special value to flag missing samples. if =100, no flagging, the function returns an error code.
   */
  double* GetData(int &aSize, const string channel_name, const double gps_start, const double gps_end, const double val_missing=100.0);

  /**
   * Prints a data vector in an text file.
   * This function prints the time series of a given channel between two GPS times in an ASCII file. 
   * @param channel_name channel name
   * @param gps_start starting GPS time
   * @param gps_end ending GPS time
   * @param outputfilename output ASCII file name
   */
  bool PrintAsciiData(const string channel_name, const double gps_start, const double gps_end, string outputfilename);

  /**
   * Plots a data vector.
   * This function, when run interactively, plots the time series of a given channel between two GPS times. A pointer to the resulting graph is returned. Do not modify or delete this object as it is internally used by the class. NULL is returned if this function fails.
   * @param channel_name channel name
   * @param gps_start starting GPS time
   * @param gps_end ending GPS time
   */
  TGraph* PlotData(const string channel_name, const double gps_start, const double gps_end);
  TGraph* PlotData2(const string channel_name1, const string channel_name12, const double gps_start, const double gps_end);

  /**
   * Plots the time-derivative of a data vector.
   * This function, when run interactively, plots the time series of the time-derivative of a given channel between two GPS times. A pointer to the resulting graph is returned. Do not modify or delete this object as it is internally used by the class. NULL is returned if this function fails.
   * @param channel_name channel name
   * @param gps_start starting GPS time
   * @param gps_end ending GPS time
   */
  TGraph* PlotDerivative(const string channel_name, const double gps_start, const double gps_end);

  /**
   * Converts the frame file list to an FFL format.
   * If the format specified with ffl() is different from the native FFL format, the frame file list is converted into a working FFL file in the tmp directory (set with DefineTmpDir()).
   *
   * The name of the converted file is returned. "none" is returned if this function fails.
   */
  string Convert2Ffl(void);

  /**
   * Extracts available channels at a given GPS time.
   * Only ADC, PROC, SIM and SER channels are extracted.
   * Only channels with a sampling rate above or equal to 1Hz are extracted.
   *
   * Note that data access is retried several times, as configured with LoadFrameFile().
   * @param gps GPS time
   */
  bool ExtractChannels(const int gps=0);

  /**
   * Returns the sampling rate of a given channel.
   * Returns 0 if the channel is not found.
   * Returns -1 if there is a error.
   * To target a given GPS time, use ExtractChannels() first.
   * @param[in] channel_name channel name
   * @param[out] channel index
   */
  int GetChannelSampling(const string channel_name, int &channel_index);

  /**
   * Returns the current channel list.
   * It is possible to return a subset of channels using a filter. This filter must be a single word containing wildcards, e.g. "V1:DQ_*_FLAG_*".
   * @param channel_filter channel name filter (substring)
   */  
  vector <string> GetChannelList(const string channel_filter="");

  /**
   * Tests whether a channel exists in the data set pointed by the ffl.
   * Returns true if the channel exists. Important: make sure that the channel list was extracted at the correct GPS time. See ExtractChannels() or LoadFrameFile().
   * @param channel_name channel name to test
   */
  bool IsChannel(const string channel_name);

  /**
   * Returns a pointer to the segments covered by the ffl.
   * DO NOT DELETE OR MODIFY! This object is internally used by the class.
   */
  inline Segments* GetSegments(void){ return seg; }

  /**
   * Resets ffl.
   */
  void Reset(void);

  /**
   * Prints available channels at a given GPS time.
   * Two columns are printed: the channel name and the sampling frequency
   * Only channels with a sampling rate above 1Hz are printed.
   * @param gps GPS time   
   */
  inline bool PrintChannels(const int gps=0){
    bool res=LoadFrameFile(gps);
    for(int l=0; l<(int)channels.size(); l++) cout<<channels[l]<<" "<<sampling[l]<<endl;
    return res;
  };
  /**
   * Returns the path to the input frame file list.
   */
  inline string GetInputFfl(void){ return fInFrameFile; };

 private:

  bool status_OK;     ///< class status
  string fInFrameFile;///< path to input frame file list
  string fWoFrameFile;///< path to working frame file (always FFL)
  string fFormat;     ///< format frame file
  int fVerbosity;     ///< verbosity level
  string tmpdir;      ///< tmpdir used to write converted ffl file
  string srandint;    ///< random integer string
  int ntries;         ///< number of tries when accessing the data
  int nsleepsec;      ///< number of seconds between tries

  vector<string> ffile;///< list of files
  Segments *seg;       ///< segments covered by the ffl file
  FrFile *frfile;      ///< Fr file
  vector <string> channels;///< list of channels
  vector <double> sampling;///< list of sampling frequencies

  // PLOTS
  GwollumPlot *GP;     ///< plotting structure
  TGraph *G;           ///< graph

  ClassDef(ffl,0)  

};

#endif


