//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __fft__
#define __fft__

#include "CUtils.h"
#include <fftw3.h>

using namespace std;


/**
 * Wrap and optimize FFTW.
 * See <a href="http://www.fftw.org/">FFTW</a>.
 * \author    Florent Robinet
 */
class fft {
  
 public:
  
  /**
   * @name Constructors and destructors
   @{
  */
  /**
   * Constructor of the fft class.
   * A FFT plan is contructed for a given time-domain input size. The input size must be an even number or else it will be forced as such.
   *
   * A FFT type must be provided. The default "c2c" type provides a complex to complex FFT. In this case, the time-domain and the frequency domain vectors are both complex and have the same size. If a "r2c" type is selected, the time-domain vector is purely real. The frequency-domain vector is complex and its size is "aSize_t/2+1".
   *
   * IMPORTANT: when using the "r2c" type, the backward transform will alter the frequency-domain vector. Do not use it after the transform. 
   *
   * Several plans are supported: "FFTW_ESTIMATE", "FFTW_MEASURE" and "FFTW_PATIENT".
   * @param aSize_t time-domain vector size: must be even
   * @param aPlan FFT plan
   * @param aType FFT type "r2c" or "c2c"
   */
  fft(const int aSize_t, const string aPlan = "FFTW_ESTIMATE", const string aType="c2c");

  /**
   * Destructor of the fft class.
   */
  virtual ~fft(void);// destructor
  /**
     @}
  */
  
  /**
   * FFT-forwards.
   */
  inline void Forward(void){ fftw_execute(FFT_Forward); };

  /**
   * FFT-backwards.
   */
  inline void Backward(void){ fftw_execute(FFT_Backward); };

  /**
   * FFT-forwards a time-domain complex data vector.
   * The input data vector is defined by 2 arrays for the real and imaginary parts. The input data vector is copied before being transformed. It must verify the size defined in the constructor.
   * This function can only be called in a "c2c" mode.
   * @param aRe_t time-domain data vector (real part)
   * @param aIm_t time-domain data vector (imaginary part)
   */
  bool Forward(double *aRe_t, double *aIm_t);

  /**
   * FFT-forwards a time-domain real data vector.
   * The input data vector is copied before being transformed. It must verify the size defined in the constructor.
   * @param aRe_t time-domain real data vector
   */
  bool Forward(double *aRe_t);

  /**
   * FFT-backwards a frequency-domain complex data vector.
   * The input data vector is defined by 2 arrays for the real and imaginary parts. The input data vector is copied before being transformed. It must verify the size defined in the constructor.
   * @param aRe_f frequency-domain data vector (real part)
   * @param aIm_f frequency-domain data vector (imaginary part)
   */
  bool Backward(double *aRe_f, double *aIm_f);

  /**
   * FFT-backwards a frequency-domain real data vector.
   * The input data vector is copied before being transformed. It must verify the size defined in the constructor.
   * @param aRe_f frequency-domain real data vector
   */
  bool Backward(double *aRe_f);

  /**
   * Returns the squared norm of a frequency-domain data vector element.
   * @param aIndex vector index
   */
  inline double GetNorm2_f(const int aIndex){
    return x_f[aIndex][0]*x_f[aIndex][0] + x_f[aIndex][1]*x_f[aIndex][1];
  };

  /**
   * Returns the squared norm of a time-domain data vector element.
   * @param aIndex vector index
   */
  inline double GetNorm2_t(const int aIndex){
    if(fType) return xc_t[aIndex][0]*xc_t[aIndex][0] + xc_t[aIndex][1]*xc_t[aIndex][1];
    return xr_t[aIndex]*xr_t[aIndex];
  };

  /**
   * Returns the norm of a frequency-domain data vector element.
   * @param aIndex vector index
   */
  inline double GetNorm_f(const int aIndex){
    return sqrt(GetNorm2_f(aIndex));
  };

  /**
   * Returns the norm of a time-domain data vector element.
   * For r2c, prefer the use of GetRe_t() (faster).
   * @param aIndex vector index
   */
  inline double GetNorm_t(const int aIndex){
    return sqrt(GetNorm2_t(aIndex));
  };

  /**
   * Returns the real part of a frequency-domain data vector element.
   * @param aIndex vector index
   */
  inline double GetRe_f(const int aIndex){
    return x_f[aIndex][0];
  };

  /**
   * Returns the real part of a time-domain data vector element.
   * @param aIndex vector index
   */
  inline double GetRe_t(const int aIndex){
    if(fType) return xc_t[aIndex][0];
    return xr_t[aIndex];
  };

  /**
   * Returns the imaginary part of a frequency-domain data vector element.
   * @param aIndex vector index
   */
  inline double GetIm_f(const int aIndex){
    return x_f[aIndex][1];
  };

  /**
   * Returns the imaginary part of a time-domain data vector element.
   * @param aIndex vector index
   */
  inline double GetIm_t(const int aIndex){
    if(fType) return xc_t[aIndex][1];
    return 0.0;
  };

  /**
   * Returns the phase of a frequency-domain data vector element.
   * @param aIndex vector index
   */
  inline double GetPhase_f(const int aIndex){
    return atan2(x_f[aIndex][1], x_f[aIndex][0]);
  };

  /**
   * Returns the phase of a time-domain data vector element.
   * @param aIndex vector index
   */
  inline double GetPhase_t(const int aIndex){
    if(fType) return atan2(xc_t[aIndex][1], xc_t[aIndex][0]);
    else return 0.0;
  };

  /**
   * Returns the squared norm of the frequency-domain data vector.
   * The user is in charge of deleting the returned array.
   * @param aNorm the returned vector is normalized by this parameter
   */
  double* GetNorm2_f(const double aNorm=1.0);

  /**
   * Returns the squared norm of the time-domain data vector.
   * The user is in charge of deleting the returned array.
   * @param aNorm the returned vector is normalized by this parameter
   */
  double* GetNorm2_t(const double aNorm=1.0);

  /**
   * Returns the norm of the frequency-domain data vector.
   * The user is in charge of deleting the returned array.
   * @param aNorm the returned vector is normalized by this parameter
   */
  double* GetNorm_f(const double aNorm=1.0);

  /**
   * Returns the norm of the time-domain data vector.
   * The user is in charge of deleting the returned array.
   * @param aNorm the returned vector is normalized by this parameter
   */
  double* GetNorm_t(const double aNorm=1.0);

  /**
   * Returns the real part of the frequency-domain data vector.
   * The user is in charge of deleting the returned array.
   * @param aNorm the returned vector is normalized by this parameter
   */
  double* GetRe_f(const double aNorm=1.0);

  /**
   * Returns the real part of the time-domain data vector.
   * The user is in charge of deleting the returned array.
   * @param aNorm the returned vector is normalized by this parameter
   */
  double* GetRe_t(const double aNorm=1.0);

  /**
   * Returns the imaginary part of the frequency-domain data vector.
   * The user is in charge of deleting the returned array.
   * @param aNorm the returned vector is normalized by this parameter
   */
  double* GetIm_f(const double aNorm=1.0);

  /**
   * Returns the imaginary part of the time-domain data vector.
   * The user is in charge of deleting the returned array.
   * @param aNorm the returned vector is normalized by this parameter
   */
  double* GetIm_t(const double aNorm=1.0);

  /**
   * Returns the phase of the frequency-domain data vector.
   * The user is in charge of deleting the returned array.
   */
  double* GetPhase_f(void);

  /**
   * Returns the phase of the time-domain data vector.
   * The user is in charge of deleting the returned array.
   */
  double* GetPhase_t(void);

  /**
   * Returns the size of the time-domain vector.
   */
  inline int GetSize_t(void){ return fSize_t; };

  /**
   * Returns the size of the frequency-domain vector.
   */
  inline int GetSize_f(void){ return fSize_f; };

  /**
   * Sets a time-domain vector element (real part) to a given value.
   * @param aIndex vector index
   * @param aValue new value
   */
  inline void SetRe_t(const int aIndex, const double aValue){
    if(fType) xc_t[aIndex][0]=aValue;
    else      xr_t[aIndex]=aValue;
  };

  /**
   * Sets a time-domain vector element (imaginary part) to a given value.
   * @param aIndex vector index
   * @param aValue new value
   */
  inline void SetIm_t(const int aIndex, const double aValue){
    xc_t[aIndex][1]=aValue;
  };

  /**
   * Sets a frequency-domain vector element (real part) to a given value.
   * @param aIndex vector index
   * @param aValue new value
   */
  inline void SetRe_f(const int aIndex, const double aValue){
    x_f[aIndex][0]=aValue;
  };

  /**
   * Sets a frequency-domain vector element (imaginary part) to a given value.
   * @param aIndex vector index
   * @param aValue new value
   */
  inline void SetIm_f(const int aIndex, const double aValue){
    x_f[aIndex][1]=aValue;
  };

  
 private:
  int fSize_t;             ///< time-domain vector size
  int fSize_f;             ///< frequency-domain vector size
  string fPlan;            ///< plan type
  bool fType;              ///< fft type 1=c2c, 0=r2c

  fftw_complex* xc_t;      ///< time-domain data - complex
  double* xr_t;            ///< time-domain data - real
  fftw_complex* x_f;       ///< frequency-domain data
  fftw_plan FFT_Forward;   ///< forward plan
  fftw_plan FFT_Backward;  ///< backward plan
};


/**
 * Wrap and optimize FFTW for the 2-dimensional case.
 * See <a href="http://www.fftw.org/">FFTW</a>.
 * \author    Florent Robinet
 */
class fft2d {
  
 public:
  
  /**
   * @name Constructors and destructors
   @{
  */
  /**
   * Constructor of the fft2d class.
   * A FFT plan is contructed for a given 2D matrix. The input sizes must be even numbers or else it will be forced as such.
   *
   * Only real-to-complex transforms are supported. As a result, the backward transform will alter the frequency-domain matrix. Do not use it after the transform.
   *
   * Several plans are supported: "FFTW_ESTIMATE", "FFTW_MEASURE" and "FFTW_PATIENT".
   * @param aSize_x spatial-domain size in the horizontal (x) direction: must be even
   * @param aSize_y spatial-domain size in the vertical (y) direction: must be even
   * @param aPlan FFT plan
   */
  fft2d(const int aSize_x, const int aSize_y, const string aPlan = "FFTW_ESTIMATE");

  /**
   * Destructor of the fft2d class.
   */
  virtual ~fft2d(void);// destructor
  /**
     @}
  */
  
  /**
   * FFT-forwards.
   */
  inline void Forward(void){ fftw_execute(FFT_Forward); };

  /**
   * FFT-backwards.
   */
  inline void Backward(void){ fftw_execute(FFT_Backward); };

  /**
   * Returns the size of the spatial-domain horizontal dimension (x).
   */
  inline int GetSize_x(void){ return fSize_x; };

  /**
   * Returns the size of the spatial-domain vertical dimension (y).
   */
  inline int GetSize_y(void){ return fSize_y; };

  /**
   * Returns the size of the frequency-domain horizontal dimension (x).
   */
  inline int GetSize_fx(void){ return fSize_fx; };

  /**
   * Returns the size of the frequency-domain vertical dimension (y).
   */
  inline int GetSize_fy(void){ return fSize_fy; };

  /**
   * Sets a spatial-domain matrix element (purely-real) to a given value.
   * By convention, aIndex_x=0, aIndex_y=0 is the bottom-left corner.
   * @param aIndex_x index for the spatial-domain horizontal dimension (x): must range from 0 to Size_x-1
   * @param aIndex_y index for the spatial-domain vertical dimension (y): must range from 0 to Size_y-1
   * @param aValue new value
   */
  inline void SetMatrixElement_xy(const int aIndex_x, const int aIndex_y, const double aValue){
    v_xy[fSize_x*(fSize_y-1-aIndex_y)+aIndex_x]=aValue;
  };

  /**
   * Sets a frequency-domain matrix element (complex number) to a given value.
   * By convention, aIndex_fx=0, aIndex_fy=0 is the bottom-left corner.
   * @param aIndex_fx index for the frequency-domain horizontal dimension (x): must range from 0 to Size_fx-1
   * @param aIndex_fy index for the frequency-domain vertical dimension (y): must range from 0 to Size_fy-1
   * @param aReValue new value (real part)
   * @param aImValue new value (imaginary part)
   */
  inline void SetMatrixElement_fxy(const int aIndex_fx, const int aIndex_fy, const double aReValue, const double aImValue=0.0){
    v_fxy[fSize_fx*(fSize_fy-1-aIndex_fy)+aIndex_fx][0]=aReValue;
    v_fxy[fSize_fx*(fSize_fy-1-aIndex_fy)+aIndex_fx][1]=aImValue;
  };

  /**
   * Returns a spatial-domain matrix element (purely-real).
   * By convention, aIndex_x=0, aIndex_y=0 is the bottom-left corner.
   * @param aIndex_x index for the spatial-domain horizontal dimension (x): must range from 0 to Size_x-1
   * @param aIndex_y index for the spatial-domain vertical dimension (y): must range from 0 to Size_y-1
   */
  inline double GetMatrixElement_xy(const int aIndex_x, const int aIndex_y){
    return v_xy[fSize_x*(fSize_y-1-aIndex_y)+aIndex_x];
  };

  /**
   * Returns a frequency-domain matrix element (real part).
   * By convention, aIndex_fx=0, aIndex_fy=0 is the bottom-left corner.
   * @param aIndex_fx index for the frequency-domain horizontal dimension (x): must range from 0 to Size_fx-1
   * @param aIndex_fy index for the frequency-domain vertical dimension (y): must range from 0 to Size_fy-1
   */
  inline double GetMatrixElementRe_fxy(const int aIndex_fx, const int aIndex_fy){
    return v_fxy[fSize_fx*(fSize_fy-1-aIndex_fy)+aIndex_fx][0];
  };

  /**
   * Returns a frequency-domain matrix element (imaginary part).
   * By convention, aIndex_fx=0, aIndex_fy=0 is the bottom-left corner.
   * @param aIndex_fx index for the frequency-domain horizontal dimension (x): must range from 0 to Size_fx-1
   * @param aIndex_fy index for the frequency-domain vertical dimension (y): must range from 0 to Size_fy-1
   */
  inline double GetMatrixElementIm_fxy(const int aIndex_fx, const int aIndex_fy){
    return v_fxy[fSize_fx*(fSize_fy-1-aIndex_fy)+aIndex_fx][1];
  };

  /**
   * Returns the squared norm of a frequency-domain matrix element.
   * @param aIndex_fx index for the frequency-domain horizontal dimension (x): must range from 0 to Size_fx-1
   * @param aIndex_fy index for the frequency-domain vertical dimension (y): must range from 0 to Size_fy-1
   */
  inline double GetNorm2_fxy(const int aIndex_fx, const int aIndex_fy){
    int ind = fSize_fx*(fSize_fy-1-aIndex_fy)+aIndex_fx;
    return v_fxy[ind][0]*v_fxy[ind][0] + v_fxy[ind][1]*v_fxy[ind][1];
  };

   /**
   * Returns the norm of a frequency-domain data matrix element.
   * @param aIndex_fx index for the frequency-domain horizontal dimension (x): must range from 0 to Size_fx-1
   * @param aIndex_fy index for the frequency-domain vertical dimension (y): must range from 0 to Size_fy-1
   */
  inline double GetNorm_fxy(const int aIndex_fx, const int aIndex_fy){
    return sqrt(GetNorm2_fxy(aIndex_fx, aIndex_fy));
  };

  /**
   * Returns the phase of a frequency-domain data matrix element.
   * @param aIndex_fx index for the frequency-domain horizontal dimension (x): must range from 0 to Size_fx-1
   * @param aIndex_fy index for the frequency-domain vertical dimension (y): must range from 0 to Size_fy-1
   */
  inline double GetPhase_fxy(const int aIndex_fx, const int aIndex_fy){
    int ind = fSize_fx*(fSize_fy-1-aIndex_fy)+aIndex_fx;
    return atan2(v_fxy[ind][1], v_fxy[ind][0]);
  };

 private:
  int fSize_x;            ///< x vector size
  int fSize_y;            ///< y vector size
  int n_xy;               ///< xy matrix size
  int n_fxy;              ///< xy frequency-domain matrix size
  int fSize_fx;           ///< frequency-domain vector size (x)
  int fSize_fy;           ///< frequency-domain vector size (y)
  string fPlan;           ///< plan type

  double* v_xy;            ///< spatial-domain data - purely real
  fftw_complex* v_fxy;     ///< frequency-domain data - complex
  fftw_plan FFT_Forward;   ///< forward plan
  fftw_plan FFT_Backward;  ///< backward plan

};


#endif
